#include "../../sf.h"

#include <sf/diag/log.h>

#include "AndroidApplication.h"
#include "AudioManager.h"
#include "jni.h"
#include "../application.h"
#include "../TextureManager.h"
#include "../../gui/widgetmanager.h"
#include "../../graphics/renderer.h"

namespace sf
{
namespace core
{

static CAndroidApplication *g_Me = 0;
static jmethodID g_GameThreadRender = 0;
static jmethodID g_GameThreadGetScreenSize = 0;
static jmethodID g_GameThreadHandleEvents = 0;
static jmethodID g_GameThreadShowLink = 0;
static jmethodID g_GameThreadStopGame = 0;

AndroidTouchSubscribeContext* CAndroidApplication::m_TouchesCallbackContext = 0;

CAndroidApplication::CAndroidApplication()
:   m_EndModal(false),
    m_Stop(false),
    m_LogicalWidth(0),
    m_LogicalHeight(0),
    m_ScreenWidth(0),
    m_ScreenHeight(0),
    m_ModalDepth(0),
    m_Paused(false)
{
    SFASSERT(!g_Me);
    g_Me = this;

    memset(m_ButtonCallbacks, 0, ButtonCallbackCount*sizeof(m_ButtonCallbacks[0]));

    SFASSERT(g_JavaVM);
    
    jclass game_thread_class = JNIGetClass("com/stargaze/sf/GameThread");
    
    g_GameThreadRender = JNIGetObjectMethod(game_thread_class, "render", "()V");
    g_GameThreadGetScreenSize = JNIGetObjectMethod(game_thread_class, "getScreenSize", "()[I");
    g_GameThreadHandleEvents = JNIGetObjectMethod(game_thread_class, "handleEvents", "()V");
    g_GameThreadShowLink = JNIGetObjectMethod(game_thread_class, "showLink", "(Ljava/lang/String;)V");
	g_GameThreadStopGame = sf::core::JNIGetObjectMethod(game_thread_class, "stopGame", "()V");

    
    jintArray jarray = (jintArray) g_JavaEnv->CallObjectMethod(g_GameThread, g_GameThreadGetScreenSize);
    jint* size = g_JavaEnv->GetIntArrayElements(jarray, 0);
    if (size)
    {
        if (size[0] > size[1])
        {
        	m_LogicalWidth = size[0];
        	m_LogicalHeight = size[1];
        }
        else
        {
        	m_LogicalWidth = size[1];
        	m_LogicalHeight = size[0];
        }
        m_ScreenWidth = m_LogicalWidth;
        m_ScreenHeight = m_LogicalHeight;
        g_JavaEnv->ReleaseIntArrayElements(jarray, size, 0);
    }
    g_JavaEnv->DeleteLocalRef(jarray);
}

CAndroidApplication::~CAndroidApplication()
{
}

bool CAndroidApplication::Init()
{
    return true;
}

void CAndroidApplication::Run()
{
    MessageCycle();
}

void CAndroidApplication::Stop()
{
	if (m_Stop) return;

	sf::core::g_JavaEnv->CallVoidMethod(sf::core::g_GameThread, g_GameThreadStopGame);
    m_Stop = true;
}

void CAndroidApplication::Shutdown()
{
}

void CAndroidApplication::Pause()
{
	if (m_Paused) return;
	m_Paused = true;

	g_TimeManager::Instance().Pause(true);
	g_AudioManager::Instance().PauseAll(true);
	if (g_Application)
		g_Application->SaveAppState();

	g_TextureManager::Instance().InvalidateSurfaces();
}

void CAndroidApplication::Resume()
{
	if (!m_Paused) return;
	m_Paused = false;

	g_TimeManager::Instance().Pause(false);
	g_AudioManager::Instance().PauseAll(false);
}

void CAndroidApplication::SetButtonCallback(ButtonCallbackTypes _type, ButtonCallback* _func)
{
	m_ButtonCallbacks[_type] = _func;
}

void CAndroidApplication::OnMenuButton()
{
	if (m_ButtonCallbacks[ButtonCallbackMenu]) return (*m_ButtonCallbacks[ButtonCallbackMenu])();
}

void CAndroidApplication::OnBackButton()
{
	if (m_ButtonCallbacks[ButtonCallbackBack]) return (*m_ButtonCallbacks[ButtonCallbackBack])();
}

void CAndroidApplication::OnSearchButton()
{
	if (m_ButtonCallbacks[ButtonCallbackSearch]) return (*m_ButtonCallbacks[ButtonCallbackSearch])();
}

void CAndroidApplication::ShowLink(const char* _link)
{
    jstring path = g_JavaEnv->NewStringUTF(_link);

    g_JavaEnv->CallVoidMethod(g_GameThread, g_GameThreadShowLink, path);

    g_JavaEnv->DeleteLocalRef(path);
}

void CAndroidApplication::DoModal()
{
    MessageCycle();
}

void CAndroidApplication::StopModal()
{
    m_EndModal = true;
}

bool CAndroidApplication::ChangeScreenMode(bool _fullscreen, bool _32bpp, size_t _width, size_t _height) 
{
    return true;
}

bool CAndroidApplication::GetOriginalDesktopDimentions(size_t &_width, size_t &_height)
{
    return true;
}

static inline void AdaptPos(float _x, float _y, float& _out_x, float& _out_y)
{
	const sf::misc::FloatVector& scale = g_Application->GetRenderer()->GetScale();
	const sf::misc::FloatVector& offset = g_Application->GetRenderer()->GetAspectOffset();
	_out_x = (_x - offset.X)/scale.X;
	_out_y = (_y - offset.Y)/scale.Y;
}

void CAndroidApplication::OnTouchBegin(int _pid, float _x, float _y)
{
	if (m_TouchesCallbackContext)
	{
		float adapted_x, adapted_y;
		AdaptPos(_x, _y, adapted_x, adapted_y);
		if (m_TouchesCallbackContext->_on_touch_begin(m_TouchesCallbackContext, _pid, adapted_x, adapted_y)) return;
	}

	if (_pid == 0)
		g_Application->GetWidgetManager().DispatchTouchBegin((int)_x, (int)_y);
}

void CAndroidApplication::OnTouchMove(int _pid, float _x, float _y)
{
	if (m_TouchesCallbackContext)
	{
		float adapted_x, adapted_y;
		AdaptPos(_x, _y, adapted_x, adapted_y);
		if (m_TouchesCallbackContext->_on_touch_move(m_TouchesCallbackContext, _pid, adapted_x, adapted_y)) return;
	}

	if (_pid == 0)
		g_Application->GetWidgetManager().DispatchTouchMove((int)_x, (int)_y);
}

void CAndroidApplication::OnTouchEnd(int _pid, float _x, float _y)
{
	if (m_TouchesCallbackContext)
	{
		float adapted_x, adapted_y;
		AdaptPos(_x, _y, adapted_x, adapted_y);
		if (m_TouchesCallbackContext->_on_touch_end(m_TouchesCallbackContext, _pid, adapted_x, adapted_y)) return;
	}

	if (_pid == 0)
		g_Application->GetWidgetManager().DispatchTouchEnd((int)_x, (int)_y);
}
	
EventInformation& CAndroidApplication::GetCurrentEvent()
{
    static EventInformation e;
    return e;
}

void CAndroidApplication::DefaultWindowProc() {}

void CAndroidApplication::GetWindowClientRect(misc::IntRect &_rc) {}

void CAndroidApplication::AdjustClientRectToWindow(misc::IntRect &_rc) {}

void CAndroidApplication::GetDesktopWindowedSpace(misc::IntRect &_rc)
{
    _rc.X = _rc.Y = 0;
    _rc.Width = m_LogicalWidth;
    _rc.Height = m_LogicalHeight;
}

void CAndroidApplication::ScreenCoordsIntoClient(misc::IntVector& _pos) {}

void CAndroidApplication::ClientCoordsIntoScreen(misc::IntVector& _pos) {}

void CAndroidApplication::SetMouseCursorPos(const misc::IntVector& _pos) {}

void CAndroidApplication::GetMouseCursorPos(misc::IntVector& _pos) {}

void CAndroidApplication::SetSysCursor(SysCursor _cursor, bool _show_now) {}

CAndroidApplication::SysCursor CAndroidApplication::GetSysCursor()
{ 
    return CursorPointer; 
}

void CAndroidApplication::ShowSysCursor(bool _show) {}

bool CAndroidApplication::IsSysCursorShown() 
{
    return false;
}

void CAndroidApplication::PublicAppDraw()
{
    AppDraw();
}

CAndroidApplication* CAndroidApplication::GetMe()
{
    return g_Me;
}

void CAndroidApplication::MessageCycle()
{
    ++m_ModalDepth;

    while (!m_Stop && !m_EndModal)
    {
    	// GameThread.handleEvents
    	g_JavaEnv->CallVoidMethod(g_GameThread, g_GameThreadHandleEvents);

    	if (m_Stop || m_EndModal) break;
    	if (m_Paused) continue;

        AppUpdate();
        // GameThread.render()
        g_JavaEnv->CallVoidMethod(g_GameThread, g_GameThreadRender);
    }

    m_EndModal = false;

   --m_ModalDepth;
}


} // core
} // sf

extern "C" void Java_com_stargaze_sf_CAndroidApplication_appDraw(JNIEnv* _env, jobject _this)
{
    SFASSERT(sf::core::g_Me);
    sf::core::g_Me->PublicAppDraw();
}

extern "C" void Java_com_stargaze_sf_CAndroidApplication_onTouchBegin(JNIEnv* _env, jobject _this, jint _pid, jfloat _x, jfloat _y)
{
    SFASSERT(sf::core::g_Me);
    sf::core::g_Me->OnTouchBegin(_pid, _x, _y);
}

extern "C" void Java_com_stargaze_sf_CAndroidApplication_onTouchMove(JNIEnv* _env, jobject _this, jint _pid, jfloat _x, jfloat _y)
{
    SFASSERT(sf::core::g_Me);
    sf::core::g_Me->OnTouchMove(_pid, _x, _y);
}

extern "C" void Java_com_stargaze_sf_CAndroidApplication_onTouchEnd(JNIEnv* _env, jobject _this, jint _pid, jfloat _x, jfloat _y)
{
    SFASSERT(sf::core::g_Me);
    sf::core::g_Me->OnTouchEnd(_pid, _x, _y);
}

extern "C" void Java_com_stargaze_sf_CAndroidApplication_stop(JNIEnv* _env, jobject _this)
{
	SFASSERT(sf::core::g_Me);
	sf::core::g_Me->Stop();
}

extern "C" void Java_com_stargaze_sf_CAndroidApplication_pause(JNIEnv* _env, jobject _this)
{
	if (sf::core::g_Me)
		sf::core::g_Me->Pause();
}

extern "C" void Java_com_stargaze_sf_CAndroidApplication_resume(JNIEnv* _env, jobject _this)
{
	if (sf::core::g_Me)
		sf::core::g_Me->Resume();
}

extern "C" void Java_com_stargaze_sf_CAndroidApplication_onMenuButton(JNIEnv* _env, jobject _this)
{
	if (sf::core::g_Me)
		sf::core::g_Me->OnMenuButton();
}

extern "C" void Java_com_stargaze_sf_CAndroidApplication_onBackButton(JNIEnv* _env, jobject _this)
{
	if (sf::core::g_Me)
		sf::core::g_Me->OnBackButton();
}

extern "C" void Java_com_stargaze_sf_CAndroidApplication_onSearchButton(JNIEnv* _env, jobject _this)
{
	if (sf::core::g_Me)
		sf::core::g_Me->OnSearchButton();
}
