#include "InAppBilling.h"

#include <sf/diag/assert.h>
#include <sf/diag/log.h>
#include <sf/core/android/jni.h>
#include <sf/core/application.h>
#include <sf/misc/sys/filewrp.h>
#include <sf/misc/unicode.h>
#include <sf/core/application.h>

namespace game
{

// jni vars
static jclass g_InAppBillingClass = 0;
static jobject g_InAppBilling = 0;
static jmethodID g_MethodIsServiceAvailable = 0;
static jmethodID g_MethodIsServiceConnected = 0;
static jmethodID g_MethodIsBillingSupported = 0;
static jmethodID g_MethodIsOnline = 0;
static jmethodID g_MethodRequestPurchase = 0;
static jmethodID g_MethodSendPurchaseDelivered = 0;
static jmethodID g_MethodRestoreTransactions = 0;

// single instance of CInAppBilling
static CInAppBilling* g_Instance = 0;
static const unsigned int MAGIC = 0x12345678;
static const int IDLEN = 100;

static std::string ConvertJavaString(JNIEnv* _env, jstring _jstr)
{
	int len = _env->GetStringLength(_jstr);

	jboolean is_copy; //unused
	const jchar* chars = _env->GetStringChars(_jstr, &is_copy);

	std::string result(chars, chars + len);
	_env->ReleaseStringChars(_jstr, chars);
	return result;
}

// callbacks from Java
extern "C" void Java_com_stargaze_billing_InAppBilling_onPurchaseEvent(
		JNIEnv* _env, jobject _this, jstring _id, jint _event)
{
	if (g_Instance)
		g_Instance->OnPurchaseEvent(
				(_id ? ConvertJavaString(_env, _id).c_str() : 0), _event);
}

// callbacks from Java
extern "C" jboolean Java_com_stargaze_billing_InAppBilling_isItemPurchased(
		JNIEnv* _env, jobject _this, jstring _path, jstring _id)
{
	return CInAppBilling::IsItemPurchased(
			sf::misc::ANSIToWString(ConvertJavaString(_env, _path)).c_str(),
			ConvertJavaString(_env, _id).c_str());
}

CInAppBilling& CInAppBilling::Instance()
{
	if (!g_Instance)
		g_Instance = new CInAppBilling;
	return *g_Instance;
}

void CInAppBilling::Load(const std::vector<std::string> & _settingsKeys)
{
	// init purchase list
	for (int i = 0; i < _settingsKeys.size(); ++i)
	{
		m_Items[sf::core::g_Application->GetSettings().GetChild("settings",
				false).GetValue(_settingsKeys[i].c_str())] = Item();
	}

	bool restore = false;

	// load saved purchases
	items_t items;
	if (Load(sf::core::g_Application->GetAppDataFolder().c_str(), items))
	{
		SFASSERT(items.size() == m_Items.size());
		if (items.size() != m_Items.size())
			return;

		// update item states
		for (items_t::const_iterator it = items.begin(), end = items.end();
				it != end; ++it)
		{
			if(m_Items.find(it->first) != m_Items.end())
			{
				Item item = it->second;

				SFLOGA("sf", sf::diag::LogInfo, "%d, %d", item.State,
					(int) item.Delivered);

				m_Items[it->first] = item;
				if (it->second.State != StatePurchased)
					restore = true;
			}
		}
	}
	else
		restore = true;

	if (restore && IsBillingSupported())
		RestoreTransactions();
}

bool CInAppBilling::Load(const wchar_t * _path, items_t & _items)
{
	wchar_t path[512];
	wcscpy(path, _path);
	wcscat(path, L"/data");

	FILE* f = sf::SFfopen(path, L"rb");

	if (f != NULL)
	{
		unsigned int magic;
		fread(&magic, sizeof(magic), 1, f);

		if (magic != MAGIC)
		{
			fclose(f);
			return false;
		}

		int count;
		fread(&count, sizeof(count), 1, f);

		for (int i = 0; i < count; ++i)
		{
			// read id length
			int idlen = 0;
			fread(&idlen, sizeof(idlen), 1, f);

			SFASSERT(idlen < IDLEN);

			// read id
			char Id[IDLEN + 1];fread
			(Id, sizeof(char), idlen, f);
			Id[idlen] = 0;

			// read item state
			int state = 0;
			fread(&state, sizeof(state), 1, f);
			if (state == StatePending)
				state = StateAvailable;

			Item item;
			item.State = (PurchaseState) state;

			// read item delivery state
			char delivered = 0;
			fread(&delivered, sizeof(delivered), 1, f);
			item.Delivered = (delivered != 0);

			_items[Id] = item;

			SFLOGA("sf", sf::diag::LogInfo, "%d, %d", item.State,
					(int) item.Delivered);
		}
		fclose(f);
		return true;
	}
	return false;
}

bool CInAppBilling::IsServiceAvailable() const
{
	return sf::core::g_JavaEnv->CallBooleanMethod(g_InAppBilling,
			g_MethodIsServiceAvailable);
}

bool CInAppBilling::IsServiceConnected() const
{
	return sf::core::g_JavaEnv->CallBooleanMethod(g_InAppBilling,
			g_MethodIsServiceConnected);
}

bool CInAppBilling::IsBillingSupported() const
{
	return sf::core::g_JavaEnv->CallBooleanMethod(g_InAppBilling,
			g_MethodIsBillingSupported);
}

bool CInAppBilling::IsOnline() const
{
	return sf::core::g_JavaEnv->CallBooleanMethod(g_InAppBilling,
			g_MethodIsOnline);
}

bool CInAppBilling::InitiatePurchase(const char * _id)
{
	items_t::const_iterator it = m_Items.find(_id);
	SFASSERT(it != m_Items.end());

	if (it != m_Items.end() && it->second.State == StateAvailable)
	{
		jstring id = sf::core::g_JavaEnv->NewStringUTF(_id);
		bool result = sf::core::g_JavaEnv->CallBooleanMethod(g_InAppBilling,
				g_MethodRequestPurchase, id);
		sf::core::g_JavaEnv->DeleteLocalRef(id);

		if (result)
			m_Items[_id].State = StatePending;
		return result;
	}
	return false;
}

void CInAppBilling::RestoreTransactions()
{
	sf::core::g_JavaEnv->CallVoidMethod(g_InAppBilling,
			g_MethodRestoreTransactions);
}

CInAppBilling::PurchaseState CInAppBilling::GetItemPurchaseState(
		const char * _id) const
{
	items_t::const_iterator it = m_Items.find(_id);
	SFASSERT(it != m_Items.end());

	return it != m_Items.end() ? it->second.State : StateAvailable;
}

bool CInAppBilling::PurchaseDelivered(const char * _id)
{
	items_t::const_iterator it = m_Items.find(_id);
	SFASSERT(it != m_Items.end());

	if (it != m_Items.end())
	{
		jstring id = sf::core::g_JavaEnv->NewStringUTF(_id);
		bool result = sf::core::g_JavaEnv->CallBooleanMethod(g_InAppBilling,
				g_MethodSendPurchaseDelivered, id);
		sf::core::g_JavaEnv->DeleteLocalRef(id);

		m_Items[_id].Delivered = true;

		Save();
		return result;
	}
	return false;
}

bool CInAppBilling::IsPurchaseDelivered(const char * _id) const
{
	items_t::const_iterator it = m_Items.find(_id);
	SFASSERT(it != m_Items.end());

	bool ret = it != m_Items.end() ? it->second.Delivered : false;
	return ret;
}

void CInAppBilling::OnPurchaseEvent(const char * _id, int _event)
{
	if (!_id)
	{
		// _id could be unknown only when there is some error in purchase process (including user cancelling)
		SFASSERT(_event == EventError);

		// error with unknown item
		// find currently pending item - error is supposed to occur with it
		bool pending_found = false;
		for (items_t::const_iterator i = m_Items.begin(), end = m_Items.end();
				i != end; ++i)
				{
			if (i->second.State == StatePending)
			{
				SFASSERT(!pending_found);
				pending_found = true;
				m_Items[i->first].State = StateAvailable;
			}
		}SFASSERT(pending_found);
		Save();
		return;
	}

	items_t::const_iterator it = m_Items.find(_id);
	if (it == m_Items.end())
	{
		// oops: non active purchase id
		return;
	}

	// error could occur only with pending items
	SFASSERT(it->second.State == StatePending || _event != EventError);

	if (_event == EventError)
	{
		SFASSERT(!it->second.Delivered);
		m_Items[_id].State = StateAvailable;
	}
	else if (_event == EventPurchase)
	{
		// if already purchased and delivered
		if (it->second.State == StatePurchased && it->second.Delivered)
		{
			// send a notification
			PurchaseDelivered(_id);
		}
		m_Items[_id].State = StatePurchased;
	}
	else if (_event == EventRefund || _event == EventCancel)
	{
		m_Items[_id].State = StateAvailable;

		// check if refunded or cancelled item was delivered
		// if it was, then DO SOMETHING!!!
		if (it->second.Delivered)
		{
			// TODO: do something
		}
	}

	Save();
}

CInAppBilling::CInAppBilling()
{
	SFASSERT(!g_Instance);

	// init jni
	if (!g_InAppBillingClass)
	{
		g_InAppBillingClass = sf::core::JNIGetClass(
				"com/stargaze/billing/InAppBilling");
		jmethodID self = sf::core::JNIGetStaticMethod(g_InAppBillingClass,
				"self", "()Lcom/stargaze/billing/InAppBilling;");
		g_InAppBilling = sf::core::g_JavaEnv->CallStaticObjectMethod(
				g_InAppBillingClass, self);
		SFASSERT(g_InAppBilling);

		g_MethodIsServiceAvailable = sf::core::JNIGetObjectMethod(
				g_InAppBillingClass, "isServiceAvailable", "()Z");
		g_MethodIsServiceConnected = sf::core::JNIGetObjectMethod(
				g_InAppBillingClass, "isServiceConnected", "()Z");
		g_MethodIsBillingSupported = sf::core::JNIGetObjectMethod(
				g_InAppBillingClass, "isBillingSupported", "()Z");
		g_MethodIsOnline = sf::core::JNIGetObjectMethod(g_InAppBillingClass,
				"isOnline", "()Z");
		g_MethodRequestPurchase = sf::core::JNIGetObjectMethod(
				g_InAppBillingClass, "requestPurchase",
				"(Ljava/lang/String;)Z");
		g_MethodSendPurchaseDelivered = sf::core::JNIGetObjectMethod(
				g_InAppBillingClass, "sendPurchaseDelivered",
				"(Ljava/lang/String;)Z");
		g_MethodRestoreTransactions = sf::core::JNIGetObjectMethod(
				g_InAppBillingClass, "restoreTransactions", "()V");
	}
}

CInAppBilling::~CInAppBilling()
{
}

void CInAppBilling::Save()
{
	wchar_t path[512];
	wcscpy(path, sf::core::g_Application->GetAppDataFolder().c_str());
	wcscat(path, L"/data");

	FILE* f = sf::SFfopen(path, L"wb");
	SFASSERT(f);

	if (f == NULL
	)
		return;

	fwrite(&MAGIC, sizeof(MAGIC), 1, f);

	int val = m_Items.size();
	fwrite(&val, sizeof(val), 1, f);

	for (items_t::const_iterator i = m_Items.begin(), end = m_Items.end();
			i != end; ++i)
			{
		// write Id
		val = i->first.length();
		fwrite(&val, sizeof(val), 1, f);
		fwrite(i->first.c_str(), sizeof(char), val, f);

		// write state
		val = i->second.State;
		fwrite(&val, sizeof(val), 1, f);

		// write delivered state
		char delivered = i->second.Delivered;
		fwrite(&delivered, sizeof(delivered), 1, f);
	}

	fclose(f);
}

bool CInAppBilling::IsItemPurchased(const wchar_t * _dataDir, const char * _id)
{
	items_t items;
	if (!Load(_dataDir, items))
		return false;

	items_t::const_iterator it = items.find(_id);
	return it != items.end() ? it->second.State == StatePurchased : false;
}

} // game
