#pragma once

#include <sf/sf.h>
#include <string>
#include <map>
#include <vector>

namespace game
{

/**
 * Singleton that implements in app purchase mechanism.
 */
class CInAppBilling
{
public:

	/**
	 * State of the purchaseable item
	 */
	enum PurchaseState
	{
		StateAvailable, // item is available for purchase
		StatePending,   // item is in the process of purchasing
		StatePurchased  // item is purchased
	};
	
	static CInAppBilling& Instance(); // creating using new, never release
	
	~CInAppBilling();

	/**
	 * Loads data about purchases; call thisfrom CApp::Init
	 * id_keys - array of purchase id keys to read from game.xml
	 */
	void Load(const std::vector<std::string> & _settingsKeys);
	
	/**
	 * Checks service availability for connection; if not, then billing is not supported at all
	 */
	bool IsServiceAvailable() const; 

	/**
	 * Is app connected to the billing service? if not and service is available, 
	 * then it is a good idea to wait for the app to connect to the service
	 */
	bool IsServiceConnected() const;

	/**
	 * Checks billing supported by installed version of Android Market
	 * billing support is discovered after the app successfully connects to the billing service
	 */
	bool IsBillingSupported() const;

	/**
	 * Checks device connection into internet
	 */
	bool IsOnline() const;
	
	/**
	 * Initates purchase process.
	 */
	bool InitiatePurchase(const char * _purchaseId);

	/**
	 * returns saved purchase state of an item
	 */
	PurchaseState GetItemPurchaseState(const char * _purchaseId) const;

	/**
	 * Reports that the app successfully handled item purchase
	 */
	bool PurchaseDelivered(const char * _purchaseId);

	/**
	 * returns true if purchase was delivered
	 */
	bool IsPurchaseDelivered(const char * _purchaseId) const;
	
	/**
	 * This method called once from Java when app connects to billing service and billing is supported
	 * if app connects before this object is constructed, this function won't be called
	 */
	void OnBillingSupported();
	
	/**
	 * This method called from Java when item's purchase state changed
	 */
	void OnPurchaseEvent(const char* _purchaseId, int _event); 
	
	/**
	 * Checks item purchase state
	 */
	static bool IsItemPurchased(const wchar_t * _dataDir, const char * _purchaseId);

private:
	struct Item
	{
		Item()
			: State(StateAvailable)
			, Delivered(false)
		{}
		
		PurchaseState State;
		bool Delivered;
	};

	typedef std::map<std::string, Item> items_t;

	CInAppBilling();
	static bool Load(const wchar_t * _path, items_t & _items);
	void Save();
	void RestoreTransactions();    

	// purchase events that come from java side of InAppBilling
	enum PurchaseEvents
	{
		EventCancel = 0, 
		EventError = 1,
		EventPurchase = 2, 
		EventRefund = 3
	};

private:
	items_t m_Items;
};

} //game
