#include "TouchesHandler.h"

#include <sf/diag/assert.h>
#include <sf/diag/log.h>
#include <sf/lua/lua.h>

namespace game
{

static FORCEINLINE int GetNearestFlagPosition(unsigned _mask)
{
	if (_mask == 0) return 0;
	if (_mask == 1) return 1;
	if (_mask == 2) return 2;
	if (_mask == 4) return 3;
	if (_mask == 8) return 4;
	if (_mask == 16) return 5;
	if (_mask == 32) return 6;
	if (_mask == 64) return 7;
	if (_mask == 128) return 8;

	SFASSERT(0);
	return -1;
}


#ifdef __SF_ANDROID
    int TouchBeginCallback(sf::core::AndroidTouchSubscribeContext *_context, int _touch, float _x, float _y)
    {
        ((CTouchesHandler*)_context->user_data)->TouchBegin(_touch, _x, _y);
        return 0;
    }

    int TouchMoveCallback(sf::core::AndroidTouchSubscribeContext *_context, int _touch, float _x, float _y)
    {
        ((CTouchesHandler*)_context->user_data)->TouchMove(_touch, _x, _y);
        return 0;
    }

    int TouchEndCallback(sf::core::AndroidTouchSubscribeContext *_context, int _touch, float _x, float _y)
    {
        ((CTouchesHandler*)_context->user_data)->TouchEnd(_touch, _x, _y);
        return 0;
    }
#endif

CTouchesHandler::CTouchesHandler()
	: m_observer(0)
	, m_TouchesCount(0)
	, m_ActiveTouchesMask(0)
{
    m_Context._on_touch_begin = &TouchBeginCallback;
   	m_Context._on_touch_move = &TouchMoveCallback;
    m_Context._on_touch_end = &TouchEndCallback;
   	m_Context.user_data = this;

    sf::core::CAndroidApplication::SubscribeToTouches(&m_Context);
}

CTouchesHandler::~CTouchesHandler()
{
	sf::core::CAndroidApplication::SubscribeToTouches(0);
}

void CTouchesHandler::Subscribe(CTouchesHandlerObserver & observer)
{
	SFASSERT(!m_observer);
	m_observer = &observer;
}

void CTouchesHandler::Unsubscribe()
{
	m_observer = 0;
}

inline static float CalcDistance(const sf::misc::IntVector& _p1, const sf::misc::IntVector& _p2)
{
    return sqrtf((float)(_p1 - _p2).ModSqr());
}

int CTouchesHandler::TouchBegin(int _touch, float _x, float _y)
{
	if (_touch >= MaxTouches) return 0;

    SFASSERT((m_ActiveTouchesMask & (1 << _touch)) == 0);
    if ((m_ActiveTouchesMask & (1 << _touch)) != 0) return 0;

    m_ActiveTouches[m_TouchesCount++] = _touch;
    m_ActiveTouchesMask |= (1 << _touch);
    m_TouchesPos[_touch].X = (int)_x;
    m_TouchesPos[_touch].Y = (int)_y;

    if (m_TouchesCount == 2)
    {
        // 2 touches
        m_PrevDistance = CalcDistance(m_TouchesPos[m_ActiveTouches[0]], m_TouchesPos[m_ActiveTouches[1]]);
    }

    return m_TouchesCount > 1;
}

int CTouchesHandler::TouchMove(int _touch, float _x, float _y)
{
	if (_touch >= MaxTouches) return 0;

    SFASSERT(m_TouchesCount);
    if (m_TouchesCount == 0) return 0;

    SFASSERT((m_ActiveTouchesMask & (1 << _touch)) != 0);
	if ((m_ActiveTouchesMask & (1 << _touch)) == 0) return 0;

    m_TouchesPos[_touch].X = (int)_x;
	m_TouchesPos[_touch].Y = (int)_y;

    if (m_TouchesCount == 1)
    {
        // trivial case
        return 0;
    }

	// if first or second touch modified
	if (m_ActiveTouches[0] == _touch || m_ActiveTouches[1] == _touch)
	{
		// processing pinch zoom
		float new_dist = CalcDistance(m_TouchesPos[m_ActiveTouches[0]], m_TouchesPos[m_ActiveTouches[1]]);
		float delta = new_dist - m_PrevDistance;
		if (delta > 5 || delta < -5)
		{
			if (m_observer && new_dist > 50 && m_PrevDistance > 50)
				m_observer->PinchZoom(new_dist/m_PrevDistance, m_TouchesPos[m_ActiveTouches[0]], m_TouchesPos[m_ActiveTouches[1]]);
			 m_PrevDistance = new_dist;
		}
	}

    return 1;
}

int CTouchesHandler::TouchEnd(int _touch, float _x, float _y)
{
	if (_touch >= MaxTouches) return 0;

    SFASSERT(m_TouchesCount);
    if (m_TouchesCount == 0) return 0;

    SFASSERT((m_ActiveTouchesMask & (1 << _touch)) != 0);
	if ((m_ActiveTouchesMask & (1 << _touch)) == 0) return 0;

	m_ActiveTouchesMask &= ~(1 << _touch);

	for (int i = 0; i < m_TouchesCount; ++i)
	{
		if (m_ActiveTouches[i] == _touch)
		{
			if (i != m_TouchesCount - 1)
			{
				// removing active touch
				memmove(m_ActiveTouches + i, m_ActiveTouches + i + 1,
						(m_TouchesCount - 1 - i)*sizeof(m_ActiveTouches[0]));
			}
			--m_TouchesCount;
			//m_ActiveTouchesMask &= ~(1 << _touch);
		    if (m_TouchesCount >= 2)
		        m_PrevDistance = CalcDistance(m_TouchesPos[m_ActiveTouches[0]], m_TouchesPos[m_ActiveTouches[1]]);
			return m_TouchesCount != 0;
		}
	}

	SFASSERT(0);
	return 0;
}


} // game

