#include <unistd.h>
#include <fcntl.h>
#include <algorithm>
#include <sf/diag/assert.h>
#include <sf/diag/log.h>

#include "jni.h"
#include "VorbisPlayer.h"

namespace sf
{
namespace core
{
namespace android
{

static ov_callbacks g_Callbacks =
{
	CVorbisPlayer::ReadFile
	, CVorbisPlayer::SeekFile
	, CVorbisPlayer::CloseFile
	, CVorbisPlayer::TellFile
};

CVorbisPlayer::CVorbisPlayer() 
	: m_FileMap(0)
	, m_Playing(false)
	, m_Pause(false)
	, m_Loop(false)
	, m_Thread(0)
{
	pthread_cond_init(&m_Cond, NULL);
	pthread_mutex_init(&m_Mutex, NULL);
}

CVorbisPlayer::~CVorbisPlayer()
{
	Close();
}

bool CVorbisPlayer::Open(const void * _file_mmap, int _length, bool _loop)
{
	SFASSERT(_file_mmap);
	SFASSERT(!m_FileMap);

	if(!_file_mmap)
		return false;

	m_FileMap = (const char*)_file_mmap;
	m_CurrentPos = m_FileMap;
	m_EndPos = m_FileMap + _length;
	m_Loop = _loop;


	if (ov_open_callbacks(this, &m_vf, NULL, 0, g_Callbacks) < 0)
	{
		SFASSERT(0);
		m_FileMap = 0;
		return false;
	}

	vorbis_info *vi = ov_info(&m_vf, -1);
	m_Track.Open(vi->rate, vi->channels, g_JavaEnv);

	return true;
}

void CVorbisPlayer::Close()
{
	if (m_FileMap)
	{
		Stop();
		m_Track.Close(g_JavaEnv);
		ov_clear(&m_vf);
		m_FileMap = 0;
	}
}

void CVorbisPlayer::Start()
{
	SFASSERT(m_FileMap);

	AutoLock lock(m_Mutex);
	if (!m_Playing)
	{
		m_Playing = true;
		m_Pause = false;
		ov_raw_seek(&m_vf, 0);
		m_Track.Play(g_JavaEnv);
		int ret = pthread_create(&m_Thread, NULL, ThreadRoutine, this);
		SFASSERT(ret == 0);
	}
	else if (m_Pause)
	{
		m_Pause = false;
		m_Track.Play(g_JavaEnv);
		pthread_cond_broadcast(&m_Cond);
	}
}

void CVorbisPlayer::Stop()
{
	SFASSERT(m_FileMap);

	{
		AutoLock lock(m_Mutex);
		if(!m_Playing)
			return;
		m_Playing = false;
		m_Track.Stop(g_JavaEnv);
		if (m_Pause)
		{
			m_Pause = false;
			pthread_cond_broadcast(&m_Cond);
		}
	}
	pthread_join(m_Thread, 0);
}

void CVorbisPlayer::Pause()
{
	SFASSERT(m_FileMap);

	AutoLock lock(m_Mutex);
	if(m_Pause || !m_Playing)
		return;
	m_Pause = true;
	m_Track.Pause(g_JavaEnv);
}

bool CVorbisPlayer::IsPlaying() const
{
	SFASSERT(m_FileMap);

	AutoLock lock(m_Mutex);
	return m_Playing && !m_Pause;
}

int CVorbisPlayer::GetDuration() const
{
	SFASSERT(m_FileMap);
	return ov_time_total(&m_vf, -1);
}

void CVorbisPlayer::SetVolume(float _vol)
{
	SFASSERT(m_FileMap);
	m_Track.SetVolume(_vol, g_JavaEnv);
}

bool CVorbisPlayer::IsValid() const
{
	return m_FileMap != 0;
}

void CVorbisPlayer::PlayRoutine(JNIEnv * _env)
{
	SFASSERT(m_FileMap);

	static const int size = 1024 * 16;
	jarray buffer = _env->NewByteArray(size);

	int current_section;

	while (true)
	{
		{
			AutoLock lock(m_Mutex);
			while (m_Playing && m_Pause)
				pthread_cond_wait(&m_Cond, &m_Mutex);
			if (!m_Playing)
				break;
		}

		void* pcmout = _env->GetPrimitiveArrayCritical(buffer, 0);
		SFASSERT(pcmout);

		long len = ov_read(&m_vf, pcmout, size, &current_section);
		_env->ReleasePrimitiveArrayCritical(buffer, pcmout, 0);

		if (len <= 0)
		{
			// end of file
			if (m_Loop)
				ov_raw_seek(&m_vf, 0);
			else
			{
				AutoLock lock(m_Mutex);
				m_Playing = false;
			}
		}
		else
			m_Track.Write(buffer, len, _env);
	}
	_env->DeleteLocalRef(buffer);
}

void * CVorbisPlayer::ThreadRoutine(void * _arg)
{
	JNIEnv* env;
	g_JavaVM->AttachCurrentThread(&env, NULL);

	env->PushLocalFrame(2);

	reinterpret_cast<CVorbisPlayer *> (_arg)->PlayRoutine(env);

	env->PopLocalFrame(NULL);
	g_JavaVM->DetachCurrentThread();

	return 0;
}

size_t CVorbisPlayer::ReadMmap(void *ptr, size_t size)
{
	if(m_CurrentPos >= m_EndPos)
		return 0; // EOF

	size_t len = std::min(size, (size_t)(m_EndPos - m_CurrentPos));
	memcpy(ptr, m_CurrentPos, len);
	m_CurrentPos += len;
	return len;
}

size_t CVorbisPlayer::SeekMmap(ogg_int64_t offset, int whence)
{
	const char * pos = m_CurrentPos;
	if (whence == SEEK_SET)
		pos = m_FileMap + offset;
	else if (whence == SEEK_END)
		pos = m_EndPos + offset;
	else
		pos += offset;

	if (pos < m_FileMap || pos > m_EndPos)
		return -1;
	m_CurrentPos = pos;
	return 0;
}

size_t CVorbisPlayer::ReadFile(void *ptr, size_t size, size_t nmemb, void *datasource)
{
	return reinterpret_cast<CVorbisPlayer *> (datasource)->ReadMmap(ptr, size * nmemb);
}

int CVorbisPlayer::SeekFile(void *datasource, ogg_int64_t offset, int whence)
{
	return reinterpret_cast<CVorbisPlayer *> (datasource)->SeekMmap(offset, whence);
}

int CVorbisPlayer::CloseFile(void *datasource)
{
	return 0;
}

long CVorbisPlayer::TellFile(void *datasource)
{
	CVorbisPlayer * p = reinterpret_cast<CVorbisPlayer *> (datasource);
	return p->m_CurrentPos - p->m_FileMap;
}

}
}
}
